// @dart=2.9

// Dart imports:
import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

// Flutter imports:
import 'package:flutter/services.dart';

// Package imports:
import 'package:http/http.dart' as http;
import 'package:logger/logger.dart';

// Project imports:
import 'package:idena_lib_dart/enums/answer_type.dart' as AnswerType;
import 'package:idena_lib_dart/enums/epoch_period.dart' as EpochPeriod;
import 'package:idena_lib_dart/enums/relevance_type.dart' as RelevantType;
import 'package:idena_lib_dart/model/dictWords.dart';
import 'package:idena_lib_dart/model/flip_examples.dart';
import 'package:idena_lib_dart/model/node_type.dart';
import 'package:idena_lib_dart/model/request/flip_get_key_request.dart';
import 'package:idena_lib_dart/model/request/flip_get_request.dart';
import 'package:idena_lib_dart/model/request/flip_longHashes_request.dart';
import 'package:idena_lib_dart/model/request/flip_shortHashes_request.dart';
import 'package:idena_lib_dart/model/request/flip_submitLongAnswers_request.dart';
import 'package:idena_lib_dart/model/request/flip_submitShortAnswers_request.dart';
import 'package:idena_lib_dart/model/request/flip_words_request.dart';
import 'package:idena_lib_dart/model/response/flip_get_key_response.dart';
import 'package:idena_lib_dart/model/response/flip_get_response.dart';
import 'package:idena_lib_dart/model/response/flip_longHashes_response.dart';
import 'package:idena_lib_dart/model/response/flip_shortHashes_response.dart';
import 'package:idena_lib_dart/model/response/flip_submitLongAnswers_response.dart';
import 'package:idena_lib_dart/model/response/flip_submitShortAnswers_response.dart';
import 'package:idena_lib_dart/model/response/flip_words_response.dart';
import 'package:idena_lib_dart/model/validation_session_infos.dart';
import 'package:idena_lib_dart/pubdev/dartssh/client.dart';
import 'package:idena_lib_dart/pubdev/dartssh/http.dart' as ssh;
import 'package:idena_lib_dart/util/crypto/rlp.dart';
import 'package:idena_lib_dart/util/crypto/utils_crypto.dart';
import 'package:idena_lib_dart/util/decrypt_flip.dart';
import 'package:idena_lib_dart/util/util_vps.dart';

class ValidationService {
  var logger = Logger();

  ValidationService({this.nodeType, this.apiUrl, this.keyApp});

  int nodeType;
  String apiUrl;
  String keyApp;

  Map<String, String> requestHeaders = {
    'Content-type': 'application/json',
  };

  String body;
  http.Response responseHttp;
  Map<String, dynamic> mapParams;
  SSHClient sshClient;

  Future<ValidationSessionInfo> getValidationSessionFlipsList(
      String typeSession,
      ValidationSessionInfo validationSessionInfoInput,
      bool simulationMode,
      String address) async {
    Completer<ValidationSessionInfo> _completer =
        new Completer<ValidationSessionInfo>();

    if (validationSessionInfoInput != null) {
      _completer.complete(validationSessionInfoInput);
      return _completer.future;
    }

    ValidationSessionInfo validationSessionInfo = new ValidationSessionInfo();
    validationSessionInfo.typeSession = typeSession;
    String method;

    switch (typeSession) {
      case EpochPeriod.ShortSession:
        {
          method = FlipShortHashesRequest.METHOD_NAME;
        }
        break;
      case EpochPeriod.LongSession:
        {
          method = FlipLongHashesRequest.METHOD_NAME;
        }
        break;
      default:
        return validationSessionInfo;
    }

    FlipShortHashesRequest flipShortHashesRequest;
    FlipShortHashesResponse flipShortHashesResponse;
    FlipLongHashesRequest flipLongHashesRequest;
    FlipLongHashesResponse flipLongHashesResponse;

    try {
      if (simulationMode) {
        if (typeSession == EpochPeriod.ShortSession) {
          flipShortHashesResponse = FlipShortHashesResponse.fromJson(
              FlipExamples().getMapExample(typeSession));
        }
        if (typeSession == EpochPeriod.LongSession) {
          flipLongHashesResponse = FlipLongHashesResponse.fromJson(
              FlipExamples().getMapExample(typeSession));
        }
      } else {
        if (this.nodeType == SHARED_NODE) {
          mapParams = {
            'method': method,
            "params": [address],
            'id': 101,
            'key': keyApp
          };
        } else {
          mapParams = {
            'method': method,
            "params": [],
            'id': 101,
            'key': keyApp
          };
        }

        if (typeSession == EpochPeriod.ShortSession) {
          if (this.nodeType == NORMAL_VPS_NODE) {
            SSHClientStatus sshClientStatus;
            sshClientStatus = await VpsUtil().connectVps(apiUrl, keyApp);
            if (sshClientStatus.sshClientStatus) {
              sshClient = sshClientStatus.sshClient;
            }
            var response = await ssh.HttpClientImpl(
                    clientFactory: () =>
                        ssh.SSHTunneledBaseClient(sshClientStatus.sshClient))
                .request(apiUrl,
                    method: 'POST',
                    data: jsonEncode(mapParams),
                    headers: requestHeaders);
            if (response.status == 200) {
              flipShortHashesResponse =
                  flipShortHashesResponseFromJson(response.text);
            }
          } else {
            flipShortHashesRequest = FlipShortHashesRequest.fromJson(mapParams);
            logger.i(new JsonEncoder.withIndent('  ')
                .convert(flipShortHashesRequest));
            body = json.encode(flipShortHashesRequest.toJson());
            responseHttp = await http.post(Uri.parse(this.apiUrl),
                body: body, headers: requestHeaders);
            if (responseHttp.statusCode == 200) {
              flipShortHashesResponse =
                  flipShortHashesResponseFromJson(responseHttp.body);
            }
          }
        }
        if (typeSession == EpochPeriod.LongSession) {
          if (this.nodeType == NORMAL_VPS_NODE) {
            SSHClientStatus sshClientStatus;
            sshClientStatus = await VpsUtil().connectVps(apiUrl, keyApp);
            if (sshClientStatus.sshClientStatus) {
              sshClient = sshClientStatus.sshClient;
            }
            var response = await ssh.HttpClientImpl(
                    clientFactory: () =>
                        ssh.SSHTunneledBaseClient(sshClientStatus.sshClient))
                .request(apiUrl,
                    method: 'POST',
                    data: jsonEncode(mapParams),
                    headers: requestHeaders);
            if (response.status == 200) {
              flipLongHashesResponse =
                  flipLongHashesResponseFromJson(response.text);
            }
          } else {
            flipLongHashesRequest = FlipLongHashesRequest.fromJson(mapParams);
            logger.i(new JsonEncoder.withIndent('  ')
                .convert(flipLongHashesRequest));
            body = json.encode(flipLongHashesRequest.toJson());
            responseHttp = await http.post(Uri.parse(this.apiUrl),
                body: body, headers: requestHeaders);
            if (responseHttp.statusCode == 200) {
              flipLongHashesResponse =
                  flipLongHashesResponseFromJson(responseHttp.body);
            }
          }
        }
      }

      List<ValidationSessionInfoFlips> listSessionValidationFlip =
          List<ValidationSessionInfoFlips>.empty(growable: true);
      List<ValidationSessionInfoFlips> listSessionValidationFlipExtra =
          List<ValidationSessionInfoFlips>.empty(growable: true);
      int nbFlips = 0;
      if (typeSession == EpochPeriod.ShortSession) {
        nbFlips = flipShortHashesResponse.result.length;
      }
      if (typeSession == EpochPeriod.LongSession) {
        nbFlips = flipLongHashesResponse.result.length;
      }

      for (int i = 0; i < nbFlips; i++) {
        ValidationSessionInfoFlips validationSessionInfoFlips =
            new ValidationSessionInfoFlips();
        if (typeSession == EpochPeriod.ShortSession) {
          validationSessionInfoFlips.hash =
              flipShortHashesResponse.result[i].hash;
          validationSessionInfoFlips.ready =
              flipShortHashesResponse.result[i].ready;
          validationSessionInfoFlips.extra =
              flipShortHashesResponse.result[i].extra;
          validationSessionInfoFlips.available =
              flipShortHashesResponse.result[i].available;
          validationSessionInfoFlips.answerType = AnswerType.NONE;
          validationSessionInfoFlips.relevanceType = RelevantType.NO_INFO;
        }
        if (typeSession == EpochPeriod.LongSession) {
          validationSessionInfoFlips.hash =
              flipLongHashesResponse.result[i].hash;
          validationSessionInfoFlips.ready =
              flipLongHashesResponse.result[i].ready;
          validationSessionInfoFlips.extra =
              flipLongHashesResponse.result[i].extra;
          validationSessionInfoFlips.available =
              flipLongHashesResponse.result[i].available;
          validationSessionInfoFlips.answerType = AnswerType.NONE;
          validationSessionInfoFlips.relevanceType = RelevantType.NO_INFO;
        }

        if (validationSessionInfoFlips.extra) {
          listSessionValidationFlipExtra.add(validationSessionInfoFlips);
        } else {
          if (this.nodeType == SHARED_NODE) {
            listSessionValidationFlip.add(validationSessionInfoFlips);
          } else {
            if (validationSessionInfoFlips.ready) {
              listSessionValidationFlip.add(validationSessionInfoFlips);
            }
          }
        }
      }

      validationSessionInfo.listSessionValidationFlips =
          listSessionValidationFlip;
      validationSessionInfo.listSessionValidationFlipsExtra =
          listSessionValidationFlipExtra;
    } catch (e) {
      logger.e(e.toString());
    }
    _completer.complete(validationSessionInfo);

    return _completer.future;
  }

  Future<ValidationSessionInfoFlips> getValidationSessionFlipDetail(
      ValidationSessionInfoFlips validationSessionInfoFlips,
      String address,
      bool simulationMode,
      String privateKey) async {
    Completer<ValidationSessionInfoFlips> _completer =
        new Completer<ValidationSessionInfoFlips>();

    try {
      // get Flip
      FlipGetResponse flipGetResponse;
      FlipGetKeyResponse flipGetKeyResponse;
      FlipGetRequest flipGetRequest;

      Uint8List imageUint8_1;
      Uint8List imageUint8_2;
      Uint8List imageUint8_3;
      Uint8List imageUint8_4;

      Decoded images;
      Decoded privateImages;
      List listImages = new List(4);
      List orders = new List(2);

      if (simulationMode) {
        String data =
            await loadAssets(validationSessionInfoFlips.hash + "_images");
        flipGetResponse = flipGetResponseFromJson(data);

        if (flipGetResponse.result.privateHex != null &&
            flipGetResponse.result.privateHex != '0x') {
          // ;[images] = decode(publicHex || hex)
          if (flipGetResponse.result.publicHex != null) {
            images = decodeRlp(
                Uint8List.fromList(toBuffer(flipGetResponse.result.publicHex)),
                true);
          } else {
            if (flipGetResponse.result.hex != null) {
              images = decodeRlp(
                  Uint8List.fromList(toBuffer(flipGetResponse.result.hex)),
                  true);
            }
          }

          // let privateImages
          // ;[privateImages, orders] = decode(privateHex)
          privateImages = decodeRlp(
              Uint8List.fromList(toBuffer(flipGetResponse.result.privateHex)),
              true);
        } else {
          // TODO: implement this case
          // ;[images, orders] = decode(hex)
          var images3 = decodeRlp(
              Uint8List.fromList(toBuffer(flipGetResponse.result.hex)), true);
        }
      } else {
        if (this.nodeType == SHARED_NODE) {
          mapParams = {
            'method': FlipGetRequest.METHOD_NAME_RAW,
            'params': [validationSessionInfoFlips.hash],
            'id': 101,
            'key': keyApp
          };

          flipGetRequest = FlipGetRequest.fromJson(mapParams);
          body = json.encode(flipGetRequest.toJson());
          responseHttp = await http.post(Uri.parse(this.apiUrl),
              body: body, headers: requestHeaders);
          if (responseHttp.statusCode == 200) {
            flipGetResponse = flipGetResponseFromJson(responseHttp.body);
          }

          mapParams = {
            'method': FlipGetKeyRequest.METHOD_NAME,
            'params': [address, validationSessionInfoFlips.hash],
            'id': 101,
            'key': keyApp
          };

          FlipGetKeyRequest flipGetKeyRequest =
              FlipGetKeyRequest.fromJson(mapParams);
          body = json.encode(flipGetKeyRequest.toJson());
          responseHttp = await http.post(Uri.parse(this.apiUrl),
              body: body, headers: requestHeaders);
          if (responseHttp.statusCode == 200) {
            flipGetKeyResponse = flipGetKeyResponseFromJson(responseHttp.body);
          }
          var decryptedPublicPart = decryptMessage(
              flipGetKeyResponse.result.publicKey,
              flipGetResponse.result.publicHex);
          var decryptedPrivateKey =
              decryptMessage(privateKey, flipGetKeyResponse.result.privateKey);
          var decryptedPrivatePart = decryptMessage(
              decryptedPrivateKey, flipGetResponse.result.privateHex);

          // const [images] = decode(decryptedPublicPart)
          images = decodeRlp(
              Uint8List.fromList(toBuffer('0x' + decryptedPublicPart)), true);

          // const [privateImages, orders] = decode(decryptedPrivatePart)
          privateImages = decodeRlp(
              Uint8List.fromList(toBuffer('0x' + decryptedPrivatePart)), true);
        } else {
          mapParams = {
            'method': FlipGetRequest.METHOD_NAME,
            'params': [validationSessionInfoFlips.hash],
            'id': 101,
            'key': keyApp
          };
          if (this.nodeType == NORMAL_VPS_NODE) {
            SSHClientStatus sshClientStatus;
            sshClientStatus = await VpsUtil().connectVps(apiUrl, keyApp);
            if (sshClientStatus.sshClientStatus) {
              sshClient = sshClientStatus.sshClient;
            }
            var response = await ssh.HttpClientImpl(
                    clientFactory: () =>
                        ssh.SSHTunneledBaseClient(sshClientStatus.sshClient))
                .request(apiUrl,
                    method: 'POST',
                    data: jsonEncode(mapParams),
                    headers: requestHeaders);
            if (response.status == 200) {
              flipGetResponse = flipGetResponseFromJson(response.text);
            }
          } else {
            flipGetRequest = FlipGetRequest.fromJson(mapParams);
            body = json.encode(flipGetRequest.toJson());
            responseHttp = await http.post(Uri.parse(this.apiUrl),
                body: body, headers: requestHeaders);
            if (responseHttp.statusCode == 200) {
              flipGetResponse = flipGetResponseFromJson(responseHttp.body);
            }
          }

          if (flipGetResponse.result.privateHex != null &&
              flipGetResponse.result.privateHex != '0x') {
            // ;[images] = decode(publicHex || hex)
            if (flipGetResponse.result.publicHex != null) {
              images = decodeRlp(
                  Uint8List.fromList(
                      toBuffer(flipGetResponse.result.publicHex)),
                  true);
            } else {
              if (flipGetResponse.result.hex != null) {
                images = decodeRlp(
                    Uint8List.fromList(toBuffer(flipGetResponse.result.hex)),
                    true);
              }
            }

            // let privateImages
            // ;[privateImages, orders] = decode(privateHex)
            privateImages = decodeRlp(
                Uint8List.fromList(toBuffer(flipGetResponse.result.privateHex)),
                true);
          } else {
            // TODO: implement this case
            // ;[images, orders] = decode(hex)
            var images3 = decodeRlp(
                Uint8List.fromList(toBuffer(flipGetResponse.result.hex)), true);
          }
        }
      }

      // images = images.concat(privateImages)
      imageUint8_1 = images.data[0][0];
      imageUint8_2 = images.data[0][1];
      imageUint8_3 = privateImages.data[0][0];
      imageUint8_4 = privateImages.data[0][1];
      listImages[0] = imageUint8_1;
      listImages[1] = imageUint8_2;
      listImages[2] = imageUint8_3;
      listImages[3] = imageUint8_4;
      orders = privateImages.data[1];

      String order1 =
          orders[0][0].toString().replaceAll('[', '').replaceAll(']', '');
      String order2 =
          orders[0][1].toString().replaceAll('[', '').replaceAll(']', '');
      String order3 =
          orders[0][2].toString().replaceAll('[', '').replaceAll(']', '');
      String order4 =
          orders[0][3].toString().replaceAll('[', '').replaceAll(']', '');
      validationSessionInfoFlips.listImagesLeft = new List<Uint8List>(4);
      validationSessionInfoFlips.listImagesLeft[0] =
          listImages[int.tryParse(order1) ?? 0];
      validationSessionInfoFlips.listImagesLeft[1] =
          listImages[int.tryParse(order2) ?? 0];
      validationSessionInfoFlips.listImagesLeft[2] =
          listImages[int.tryParse(order3) ?? 0];
      validationSessionInfoFlips.listImagesLeft[3] =
          listImages[int.tryParse(order4) ?? 0];

      // TODO .. dirty
      order1 = orders[1][0].toString().replaceAll('[', '').replaceAll(']', '');
      order2 = orders[1][1].toString().replaceAll('[', '').replaceAll(']', '');
      order3 = orders[1][2].toString().replaceAll('[', '').replaceAll(']', '');
      order4 = orders[1][3].toString().replaceAll('[', '').replaceAll(']', '');
      validationSessionInfoFlips.listImagesRight = new List<Uint8List>(4);
      validationSessionInfoFlips.listImagesRight[0] =
          listImages[int.tryParse(order1) ?? 0];
      validationSessionInfoFlips.listImagesRight[1] =
          listImages[int.tryParse(order2) ?? 0];
      validationSessionInfoFlips.listImagesRight[2] =
          listImages[int.tryParse(order3) ?? 0];
      validationSessionInfoFlips.listImagesRight[3] =
          listImages[int.tryParse(order4) ?? 0];
    } catch (e) {
      logger.e(e.toString());
    }

    //print("flip loaded : " + validationSessionInfoFlips.hash);

    _completer.complete(validationSessionInfoFlips);
    return _completer.future;
  }

  Future<List<Word>> getWordsFromHash(
      String hash, bool simulationMode, List<Word> wordsMap) async {
    FlipWordsResponse flipWordsResponse;
    Completer<List<Word>> _completer = new Completer<List<Word>>();

    int nbWords = 0;
    if (simulationMode) {
      try {
        String data = await loadAssets(hash + "_words");
        flipWordsResponse = flipWordsResponseFromJson(data);
      } catch (e) {
        Map<String, dynamic> mapWords = {
          "jsonrpc": "2.0",
          "id": 51,
          "result": {
            "words": [0, 0]
          }
        };
        flipWordsResponse = FlipWordsResponse.fromJson(mapWords);
      }
      nbWords = flipWordsResponse.result.words.length;
    } else {
      try {
        mapParams = {
          'method': FlipWordsRequest.METHOD_NAME,
          'params': [hash],
          'id': 101,
          'key': keyApp
        };

        if (this.nodeType == NORMAL_VPS_NODE) {
          SSHClientStatus sshClientStatus;
          sshClientStatus = await VpsUtil().connectVps(apiUrl, keyApp);
          if (sshClientStatus.sshClientStatus) {
            sshClient = sshClientStatus.sshClient;
          }
          var response = await ssh.HttpClientImpl(
              clientFactory: () =>
                  ssh.SSHTunneledBaseClient(sshClientStatus.sshClient)).request(
              apiUrl,
              method: 'POST',
              data: jsonEncode(mapParams),
              headers: requestHeaders);
          if (response.status == 200) {
            flipWordsResponse = flipWordsResponseFromJson(response.text);
            nbWords = flipWordsResponse.result.words.length;
          }
        } else {
          FlipWordsRequest flipWordsRequest =
              FlipWordsRequest.fromJson(mapParams);
          body = json.encode(flipWordsRequest.toJson());
          responseHttp = await http.post(Uri.parse(this.apiUrl),
              body: body, headers: requestHeaders);
          if (responseHttp.statusCode == 200) {
            flipWordsResponse = flipWordsResponseFromJson(responseHttp.body);
            nbWords = flipWordsResponse.result.words.length;
          }
        }
      } catch (e) {}
    }
    List<Word> listWords = new List(nbWords);
    for (int j = 0; j < nbWords; j++) {
      Word word;
      if (flipWordsResponse.result.words[j] == 0) {
        word = new Word(name: "", desc: "");
      } else {
        word = new Word(
            name: wordsMap[flipWordsResponse.result.words[j]].name,
            desc: wordsMap[flipWordsResponse.result.words[j]].desc);
      }
      listWords[j] = word;
    }

    _completer.complete(listWords);
    return _completer.future;
  }

  Future<String> loadAssets(String fileName) async {
    try {
      return await rootBundle.loadString('test/examples/' + fileName + '.json',
          cache: false);
    } catch (e) {
      return null;
    }
  }

  Future<FlipSubmitShortAnswersResponse> submitShortAnswers(
      ValidationSessionInfo validationSessionInfo) async {
    Completer<FlipSubmitShortAnswersResponse> _completer =
        new Completer<FlipSubmitShortAnswersResponse>();

    if (validationSessionInfo == null) {
      _completer.complete(null);
      return _completer.future;
    }

    FlipSubmitShortAnswersRequest flipSubmitShortAnswersRequest =
        new FlipSubmitShortAnswersRequest();
    FlipSubmitShortAnswersResponse flipSubmitShortAnswersResponse;

    try {
      ParamShortAnswer answers = new ParamShortAnswer();
      List<ShortAnswer> listAnswers = List<ShortAnswer>.empty(growable: true);
      for (int i = 0;
          i < validationSessionInfo.listSessionValidationFlips.length;
          i++) {
        ShortAnswer answer = new ShortAnswer(
            answer:
                validationSessionInfo.listSessionValidationFlips[i].answerType,
            hash: validationSessionInfo.listSessionValidationFlips[i].hash);
        listAnswers.add(answer);
      }

      answers.epoch = 0;
      answers.nonce = 0;
      answers.answers = listAnswers;

      List<ParamShortAnswer> params =
          List<ParamShortAnswer>.empty(growable: true);
      params.add(answers);
      flipSubmitShortAnswersRequest.method =
          FlipSubmitShortAnswersRequest.METHOD_NAME;
      flipSubmitShortAnswersRequest.params = params;
      flipSubmitShortAnswersRequest.id = 101;
      flipSubmitShortAnswersRequest.key = keyApp;

      if (this.nodeType == NORMAL_VPS_NODE) {
        SSHClientStatus sshClientStatus;
        sshClientStatus = await VpsUtil().connectVps(apiUrl, keyApp);
        if (sshClientStatus.sshClientStatus) {
          sshClient = sshClientStatus.sshClient;
        }
        var response = await ssh.HttpClientImpl(
            clientFactory: () =>
                ssh.SSHTunneledBaseClient(sshClientStatus.sshClient)).request(
            apiUrl,
            method: 'POST',
            data: jsonEncode(mapParams),
            headers: requestHeaders);
        if (response.status == 200) {
          flipSubmitShortAnswersResponse =
              flipSubmitShortAnswersResponseFromJson(response.text);
        }
      } else {
        body = json.encode(flipSubmitShortAnswersRequest.toJson());
        logger.i(new JsonEncoder.withIndent('  ')
            .convert(flipSubmitShortAnswersRequest));
        responseHttp = await http.post(Uri.parse(this.apiUrl),
            body: body, headers: requestHeaders);
        if (responseHttp.statusCode == 200) {
          flipSubmitShortAnswersResponse =
              flipSubmitShortAnswersResponseFromJson(responseHttp.body);
        }
      }
    } catch (e) {
      logger.e(e.toString());
    }
    _completer.complete(flipSubmitShortAnswersResponse);
    return _completer.future;
  }

  Future<FlipSubmitLongAnswersResponse> submitLongAnswers(
      ValidationSessionInfo validationSessionInfo) async {
    Completer<FlipSubmitLongAnswersResponse> _completer =
        new Completer<FlipSubmitLongAnswersResponse>();

    if (validationSessionInfo == null) {
      _completer.complete(null);
      return _completer.future;
    }
    FlipSubmitLongAnswersRequest flipSubmitLongAnswersRequest =
        new FlipSubmitLongAnswersRequest();
    FlipSubmitLongAnswersResponse flipSubmitLongAnswersResponse;
    bool wrongWordsBool;

    try {
      ParamLongAnswer answers = new ParamLongAnswer();
      List<LongAnswer> listAnswers = List<LongAnswer>.empty(growable: true);
      for (int i = 0;
          i < validationSessionInfo.listSessionValidationFlips.length;
          i++) {
        wrongWordsBool = true;
        if (validationSessionInfo.listSessionValidationFlips[i] != null &&
            validationSessionInfo.listSessionValidationFlips[i].relevanceType ==
                RelevantType.RELEVANT) {
          wrongWordsBool = false;
        }
        LongAnswer answer = new LongAnswer(
            answer:
                validationSessionInfo.listSessionValidationFlips[i].answerType,
            wrongWords: wrongWordsBool,
            hash: validationSessionInfo.listSessionValidationFlips[i].hash);
        listAnswers.add(answer);
      }

      answers.epoch = 0;
      answers.nonce = 0;
      answers.answers = listAnswers;

      List<ParamLongAnswer> params =
          List<ParamLongAnswer>.empty(growable: true);
      params.add(answers);
      flipSubmitLongAnswersRequest.method =
          FlipSubmitLongAnswersRequest.METHOD_NAME;
      flipSubmitLongAnswersRequest.params = params;
      flipSubmitLongAnswersRequest.id = 101;
      flipSubmitLongAnswersRequest.key = keyApp;

      if (this.nodeType == NORMAL_VPS_NODE) {
        SSHClientStatus sshClientStatus;
        sshClientStatus = await VpsUtil().connectVps(apiUrl, keyApp);
        if (sshClientStatus.sshClientStatus) {
          sshClient = sshClientStatus.sshClient;
        }
        var response = await ssh.HttpClientImpl(
            clientFactory: () =>
                ssh.SSHTunneledBaseClient(sshClientStatus.sshClient)).request(
            apiUrl,
            method: 'POST',
            data: jsonEncode(mapParams),
            headers: requestHeaders);
        if (response.status == 200) {
          flipSubmitLongAnswersResponse =
              flipSubmitLongAnswersResponseFromJson(response.text);
        }
      } else {
        body = json.encode(flipSubmitLongAnswersRequest.toJson());
        logger.i(new JsonEncoder.withIndent('  ')
            .convert(flipSubmitLongAnswersRequest));
        responseHttp = await http.post(Uri.parse(this.apiUrl),
            body: body, headers: requestHeaders);
        if (responseHttp.statusCode == 200) {
          flipSubmitLongAnswersResponse =
              flipSubmitLongAnswersResponseFromJson(responseHttp.body);
        }
      }
    } catch (e) {
      logger.e(e.toString());
    }
    _completer.complete(flipSubmitLongAnswersResponse);
    return _completer.future;
  }
}
