// @dart=2.9
// To parse this JSON data, do
//
//     final bcnTransactionsResponse = bcnTransactionsResponseFromJson(jsonString);

// Dart imports:
import 'dart:convert';

// Project imports:
import 'package:idena_lib_dart/model/address.dart';
import 'package:idena_lib_dart/model/smartContractMultiSig.dart';
import 'package:idena_lib_dart/util/util_number.dart';

BcnTransactionsResponse bcnTransactionsResponseFromJson(String str) =>
    BcnTransactionsResponse.fromJson(json.decode(str));

String bcnTransactionsResponseToJson(BcnTransactionsResponse data) =>
    json.encode(data.toJson());

class BcnTransactionsResponse {
  BcnTransactionsResponse({
    this.jsonrpc,
    this.id,
    this.result,
  });

  String jsonrpc;
  int id;
  BcnTransactionsResponseResult result;

  factory BcnTransactionsResponse.fromJson(Map<String, dynamic> json) =>
      BcnTransactionsResponse(
        jsonrpc: json["jsonrpc"],
        id: json["id"],
        result: BcnTransactionsResponseResult.fromJson(json["result"]),
      );

  Map<String, dynamic> toJson() => {
        "jsonrpc": jsonrpc,
        "id": id,
        "result": result.toJson(),
      };
}

class BcnTransactionsResponseResult {
  BcnTransactionsResponseResult({
    this.transactions,
  });

  List<Transaction> transactions;

  factory BcnTransactionsResponseResult.fromJson(Map<String, dynamic> json) =>
      BcnTransactionsResponseResult(
        transactions: json["transactions"] == null
            ? null
            : List<Transaction>.from(
                json["transactions"].map((x) => Transaction.fromJson(x))),
      );

  Map<String, dynamic> toJson() => {
        "transactions": List<dynamic>.from(transactions.map((x) => x.toJson())),
      };
}

class Transaction {
  Transaction({
    this.hash,
    this.type,
    this.from,
    this.to,
    this.amount,
    this.tips,
    this.maxFee,
    this.nonce,
    this.epoch,
    this.payload,
    this.blockHash,
    this.usedFee,
    this.timestamp,
    this.smartContractMultiSig,
  });

  String hash;
  String type;
  String from;
  dynamic to;
  String amount;
  String tips;
  String maxFee;
  int nonce;
  int epoch;
  String payload;
  String blockHash;
  String usedFee;
  int timestamp;
  SmartContractMultiSig smartContractMultiSig;

  factory Transaction.fromJson(Map<String, dynamic> json) => Transaction(
        hash: json["hash"],
        type: json["type"],
        from: json["from"],
        to: json["to"],
        amount: json["amount"],
        tips: json["tips"],
        maxFee: json["maxFee"],
        nonce: json["nonce"],
        epoch: json["epoch"],
        payload: json["payload"],
        blockHash: json["blockHash"],
        usedFee: json["usedFee"],
        timestamp: json["timestamp"],
      );

  Map<String, dynamic> toJson() => {
        "hash": hash,
        "type": type,
        "from": from,
        "to": to,
        "amount": amount,
        "tips": tips,
        "maxFee": maxFee,
        "nonce": nonce,
        "epoch": epoch,
        "payload": payload,
        "blockHash": blockHash,
        "usedFee": usedFee,
        "timestamp": timestamp,
      };

  String getFormattedAmount() {
    return NumberUtil.getRawAsUsableString(amount.toString());
  }

  String getShortString(String selectedAccountAddress) {
    if (this.from != selectedAccountAddress) {
      return new Address(this.from).getShorterString();
    } else {
      return new Address(this.to).getShorterString();
    }
  }

  String getShorterString(String selectedAccountAddress) {
    if (this.from != selectedAccountAddress) {
      return new Address(this.from).getShorterString();
    } else {
      return new Address(this.to).getShorterString();
    }
  }
}
